from math import ceil, log

import torch
import torch.nn as nn
import torch.nn.functional as F

from .model import Discriminator, Generator

__all__ = ["AutoEncodingGenerator", "AutoEncodingDiscriminator"]


class AutoEncodingGenerator(Generator):
    r"""Autoencoding Generator for Boundary Equilibrium GAN (BEGAN) from
    `"BEGAN : Boundary Equilibrium Generative Adversarial Networks
    by Berthelot et. al." <https://arxiv.org/abs/1703.10717>`_ paper

    Args:
        encoding_dims (int, optional): Dimension of the encoding vector sampled from the noise prior.
        out_size (int, optional): Height and width of the input image to be generated. Must be at
            least 16 and should be an exact power of 2.
        out_channels (int, optional): Number of channels in the output Tensor.
        step_channels (int, optional): Number of channels in multiples of which the DCGAN steps up
            the convolutional features. The step up is done as dim :math:`z \rightarrow d \rightarrow
            2 \times d \rightarrow 4 \times d \rightarrow 8 \times d` where :math:`d` = step_channels.
        scale_factor (int, optional): The scale factor is used to infer properties of the model like
            ``upsample_pad``, ``upsample_filters``, ``upsample_stride`` and ``upsample_output_pad``.
        batchnorm (bool, optional): If True, use batch normalization in the convolutional layers of
            the generator.
        nonlinearity (torch.nn.Module, optional): Nonlinearity to be used in the intermediate
            convolutional layers. Defaults to ``LeakyReLU(0.2)`` when None is passed.
        last_nonlinearity (torch.nn.Module, optional): Nonlinearity to be used in the final
            convolutional layer. Defaults to ``Tanh()`` when None is passed.
        label_type (str, optional): The type of labels expected by the Generator. The available
            choices are 'none' if no label is needed, 'required' if the original labels are
            needed and 'generated' if labels are to be sampled from a distribution.
    """

    def __init__(
        self,
        encoding_dims=100,
        out_size=32,
        out_channels=3,
        step_channels=64,
        scale_factor=2,
        batchnorm=True,
        nonlinearity=None,
        last_nonlinearity=None,
        label_type="none",
    ):
        super(AutoEncodingGenerator, self).__init__(encoding_dims, label_type)
        if out_size < (scale_factor ** 4) or ceil(log(out_size, scale_factor)) != log(
            out_size, scale_factor
        ):
            raise Exception(
                "Target image size must be at least {} and a perfect power of {}".format(
                    scale_factor ** 4, scale_factor
                )
            )
        num_repeats = int(log(out_size, scale_factor)) - 3
        same_filters = scale_factor + 1
        same_pad = scale_factor // 2
        if scale_factor == 2:
            upsample_filters = 3
            upsample_stride = 2
            upsample_pad = 1
            upsample_output_pad = 1
        else:
            upsample_filters = scale_factor
            upsample_stride = scale_factor
            upsample_pad = 0
            upsample_output_pad = 0
        self.ch = out_channels
        self.n = step_channels
        use_bias = not batchnorm
        nl = nn.ELU() if nonlinearity is None else nonlinearity
        last_nl = nn.Tanh() if last_nonlinearity is None else last_nonlinearity
        init_dim = scale_factor ** 3
        self.init_dim = init_dim

        if batchnorm is True:
            self.fc = nn.Sequential(
                nn.Linear(self.encoding_dims, (init_dim ** 2) * self.n),
                nn.BatchNorm1d((init_dim ** 2) * self.n),
                nl,
            )
            initial_unit = nn.Sequential(
                nn.Conv2d(self.n, self.n, same_filters, 1, same_pad, bias=use_bias),
                nn.BatchNorm2d(self.n),
                nl,
                nn.Conv2d(self.n, self.n, same_filters, 1, same_pad, bias=use_bias),
                nn.BatchNorm2d(self.n),
                nl,
            )
            upsample_unit = nn.Sequential(
                nn.ConvTranspose2d(
                    self.n,
                    self.n,
                    upsample_filters,
                    upsample_stride,
                    upsample_pad,
                    upsample_output_pad,
                    bias=use_bias,
                ),
                nn.BatchNorm2d(self.n),
                nl,
                nn.Conv2d(self.n, self.n, same_filters, 1, same_pad, bias=use_bias),
                nn.BatchNorm2d(self.n),
                nl,
            )
        else:
            self.fc = nn.Sequential(
                nn.Linear(self.encoding_dims, (init_dim ** 2) * self.n), nl
            )
            initial_unit = nn.Sequential(
                nn.Conv2d(self.n, self.n, same_filters, 1, same_pad, bias=use_bias),
                nl,
                nn.Conv2d(self.n, self.n, same_filters, 1, same_pad, bias=use_bias),
                nl,
            )
            upsample_unit = nn.Sequential(
                nn.ConvTranspose2d(
                    self.n,
                    self.n,
                    upsample_filters,
                    upsample_stride,
                    upsample_pad,
                    upsample_output_pad,
                    bias=use_bias,
                ),
                nl,
                nn.Conv2d(self.n, self.n, same_filters, 1, same_pad, bias=use_bias),
                nl,
            )

        last_unit = nn.Sequential(
            nn.Conv2d(self.n, self.ch, same_filters, 1, same_pad, bias=True), last_nl
        )
        model = [initial_unit]
        for i in range(num_repeats):
            model.append(upsample_unit)
            out_size = out_size // scale_factor
        model.append(last_unit)
        self.model = nn.Sequential(*model)
        self._weight_initializer()

    def forward(self, z):
        r"""Calculates the output tensor on passing the encoding ``z`` through the Generator.

        Args:
            z (torch.Tensor): A 2D torch tensor of the encoding sampled from a probability
                distribution.

        Returns:
            A 4D torch.Tensor of the generated image.
        """
        x = self.fc(z)
        x = x.view(-1, self.n, self.init_dim, self.init_dim)
        return self.model(x)


class AutoEncodingDiscriminator(Discriminator):
    r"""Autoencoding Generator for Boundary Equilibrium GAN (BEGAN) from
    `"BEGAN : Boundary Equilibrium Generative Adversarial Networks
    by Berthelot et. al." <https://arxiv.org/abs/1703.10717>`_ paper

    Args:
        in_size (int, optional): Height and width of the input image to be evaluated. Must be at
            least 16 and should be an exact power of 2.
        in_channels (int, optional): Number of channels in the input Tensor.
        step_channels (int, optional): Number of channels in multiples of which the DCGAN steps up
            the convolutional features. The step up is done as dim :math:`z \rightarrow d \rightarrow
            2 \times d \rightarrow 4 \times d \rightarrow 8 \times d` where :math:`d` = step_channels.
        scale_factor (int, optional): The scale factor is used to infer properties of the model like
            ``downsample_pad``, ``downsample_filters`` and ``downsample_stride``.
        batchnorm (bool, optional): If True, use batch normalization in the convolutional layers of
            the generator.
        nonlinearity (torch.nn.Module, optional): Nonlinearity to be used in the intermediate
            convolutional layers. Defaults to ``LeakyReLU(0.2)`` when None is passed.
        last_nonlinearity (torch.nn.Module, optional): Nonlinearity to be used in the final
            convolutional layer. Defaults to ``Tanh()`` when None is passed.
        energy (bool, optional) : If set to True returns the energy instead of the decoder output.
        embeddings (bool, optional) : If set to True the embeddings will be returned.
        label_type (str, optional): The type of labels expected by the Generator. The available
            choices are 'none' if no label is needed, 'required' if the original labels are
            needed and 'generated' if labels are to be sampled from a distribution.
    """

    def __init__(
        self,
        in_size=32,
        in_channels=3,
        encoding_dims=100,
        step_channels=64,
        scale_factor=2,
        batchnorm=True,
        nonlinearity=None,
        last_nonlinearity=None,
        energy=True,
        embeddings=False,
        label_type="none",
    ):
        super(AutoEncodingDiscriminator, self).__init__(in_channels, label_type)
        if in_size < (scale_factor ** 4) or ceil(log(in_size, scale_factor)) != log(
            in_size, scale_factor
        ):
            raise Exception(
                "Input image size must be at least {} and a perfect power of {}".format(
                    scale_factor ** 4, scale_factor
                )
            )
        num_repeats = int(log(in_size, scale_factor)) - 3
        same_filters = scale_factor + 1
        same_pad = scale_factor // 2
        if scale_factor == 2:
            downsample_filters = 3
            downsample_stride = 2
            downsample_pad = 1
        else:
            downsample_filters = scale_factor
            downsample_stride = scale_factor
            downsample_pad = 0
        self.n = step_channels
        nl = nn.ELU() if nonlinearity is None else nonlinearity
        last_nl = nn.ELU() if last_nonlinearity is None else last_nonlinearity
        use_bias = not batchnorm
        init_dim = scale_factor ** 3
        self.init_dim = init_dim
        model = []
        model.append(
            nn.Sequential(
                nn.Conv2d(
                    self.input_dims, self.n, same_filters, 1, same_pad, bias=True
                ),
                nl,
            )
        )
        if batchnorm is True:
            for i in range(1, num_repeats + 1):
                model.append(
                    nn.Sequential(
                        nn.Conv2d(
                            self.n * i,
                            self.n * i,
                            same_filters,
                            1,
                            same_pad,
                            bias=use_bias,
                        ),
                        nn.BatchNorm2d(self.n * i),
                        nl,
                        nn.Conv2d(
                            self.n * i,
                            self.n * (i + 1),
                            downsample_filters,
                            downsample_stride,
                            downsample_pad,
                            bias=use_bias,
                        ),
                        nn.BatchNorm2d(self.n * (i + 1)),
                        nl,
                    )
                )
            model.append(
                nn.Sequential(
                    nn.Conv2d(
                        self.n * (num_repeats + 1),
                        self.n * (num_repeats + 1),
                        same_filters,
                        1,
                        same_pad,
                        bias=use_bias,
                    ),
                    nn.BatchNorm2d(self.n * (num_repeats + 1)),
                    nl,
                    nn.Conv2d(
                        self.n * (num_repeats + 1),
                        self.n * (num_repeats + 1),
                        same_filters,
                        1,
                        same_pad,
                        bias=use_bias,
                    ),
                    nn.BatchNorm2d(self.n * (num_repeats + 1)),
                    nl,
                )
            )
            self.fc = nn.Sequential(
                nn.Linear((init_dim ** 2) * (num_repeats + 1) * self.n, encoding_dims),
                nn.BatchNorm1d(encoding_dims),
                last_nl,
            )
        else:
            for i in range(1, num_repeats + 1):
                model.append(
                    nn.Sequential(
                        nn.Conv2d(self.n * i, self.n * i, 3, 1, 1, bias=use_bias),
                        nl,
                        nn.Conv2d(
                            self.n * i,
                            self.n * (i + 1),
                            downsample_filters,
                            downsample_stride,
                            downsample_pad,
                            bias=use_bias,
                        ),
                        nl,
                    )
                )
            model.append(
                nn.Sequential(
                    nn.Conv2d(
                        self.n * (num_repeats + 1),
                        self.n * (num_repeats + 1),
                        3,
                        1,
                        1,
                        bias=use_bias,
                    ),
                    nl,
                    nn.Conv2d(
                        self.n * (num_repeats + 1),
                        self.n * (num_repeats + 1),
                        3,
                        1,
                        1,
                        bias=use_bias,
                    ),
                    nl,
                )
            )
            self.fc = nn.Sequential(
                nn.Linear((init_dim ** 2) * (num_repeats + 1) * self.n, encoding_dims),
                last_nl,
            )
        self.encoder = nn.Sequential(*model)
        self.decoder = AutoEncodingGenerator(
            encoding_dims,
            in_size,
            in_channels,
            step_channels,
            scale_factor,
            batchnorm,
            nonlinearity,
            last_nonlinearity,
        )
        self.energy = energy
        self.embeddings = embeddings
        self._weight_initializer()

    def forward(self, x, feature_matching=False):
        r"""Calculates the output tensor on passing the image ``x`` through the Discriminator.

        Args:
            x (torch.Tensor): A 4D torch tensor of the image.
            feature_matching (bool, optional): Returns the activation from a predefined intermediate
                layer.

        Returns:
            A 1D torch.Tensor of the energy value of each image.
        """
        x1 = self.encoder(x)
        x2 = x1.view(-1, (self.init_dim ** 2) * x1.size(1))
        x2 = self.fc(x2)
        if feature_matching is True:
            return x2
        x2 = self.decoder(x2)
        if self.energy:
            x = x.view(-1, x.size(1) * x.size(2) * x.size(3))
            x2 = x2.view(-1, x2.size(1) * x2.size(2) * x2.size(3))
            if self.embeddings:
                return x1, torch.mean((x - x2) ** 2, 1)
            else:
                return torch.mean((x - x2) ** 2, 1)
        else:
            if self.embeddings:
                return x1, x2
            else:
                return x2



import torch
import torch.nn as nn

__all__ = ["GeneratorLoss", "DiscriminatorLoss"]


class GeneratorLoss(nn.Module):
    r"""Base class for all generator losses.

    .. note:: All Losses meant to be minimized for optimizing the Generator must subclass this.

    Args:
        reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the outputs are averaged over batch size.
            If ``sum`` the elements of the output are summed.
        override_train_ops (function, optional): Function to be used in place of the default ``train_ops``
    """

    def __init__(self, reduction="mean", override_train_ops=None):
        super(GeneratorLoss, self).__init__()
        self.reduction = reduction
        self.override_train_ops = override_train_ops
        self.arg_map = {}

    def set_arg_map(self, value):
        r"""Updates the ``arg_map`` for passing a different value to the ``train_ops``.

        Args:
            value (dict): A mapping of the ``argument name`` in the method signature and the
                variable name in the ``Trainer`` it corresponds to.

        .. note::
            If the ``train_ops`` signature is
            ``train_ops(self, gen, disc, optimizer_generator, device, batch_size, labels=None)``
            then we need to map ``gen`` to ``generator`` and ``disc`` to ``discriminator``.
            In this case we make the following function call
            ``loss.set_arg_map({"gen": "generator", "disc": "discriminator"})``.
        """
        self.arg_map.update(value)

    def train_ops(
        self,
        generator,
        discriminator,
        optimizer_generator,
        device,
        batch_size,
        labels=None,
    ):
        r"""Defines the standard ``train_ops`` used by most losses. Losses which have a different
        training procedure can either ``subclass`` it **(recommended approach)** or make use of
        ``override_train_ops`` argument.

        The ``standard optimization algorithm`` for the ``generator`` defined in this train_ops
        is as follows:

        1. :math:`fake = generator(noise)`
        2. :math:`value = discriminator(fake)`
        3. :math:`loss = loss\_function(value)`
        4. Backpropagate by computing :math:`\nabla loss`
        5. Run a step of the optimizer for generator

        Args:
            generator (torchgan.models.Generator): The model to be optimized.
            discriminator (torchgan.models.Discriminator): The discriminator which judges the
                performance of the generator.
            optimizer_generator (torch.optim.Optimizer): Optimizer which updates the ``parameters``
                of the ``generator``.
            device (torch.device): Device on which the ``generator`` and ``discriminator`` is present.
            batch_size (int): Batch Size of the data infered from the ``DataLoader`` by the ``Trainer``.
            labels (torch.Tensor, optional): Labels for the data.

        Returns:
            Scalar value of the loss.
        """
        if self.override_train_ops is not None:
            return self.override_train_ops(
                generator,
                discriminator,
                optimizer_generator,
                device,
                batch_size,
                labels,
            )
        else:
            if labels is None and generator.label_type == "required":
                raise Exception("GAN model requires labels for training")
            noise = torch.randn(batch_size, generator.encoding_dims, device=device)
            optimizer_generator.zero_grad()
            if generator.label_type == "generated":
                label_gen = torch.randint(
                    0, generator.num_classes, (batch_size,), device=device
                )
            if generator.label_type == "none":
                fake = generator(noise)
            elif generator.label_type == "required":
                fake = generator(noise, labels)
            elif generator.label_type == "generated":
                fake = generator(noise, label_gen)
            if discriminator.label_type == "none":
                dgz = discriminator(fake)
            else:
                if generator.label_type == "generated":
                    dgz = discriminator(fake, label_gen)
                else:
                    dgz = discriminator(fake, labels)
            loss = self.forward(dgz)
            loss.backward()
            optimizer_generator.step()
            # NOTE(avik-pal): This will error if reduction is is 'none'
            return loss.item()


class DiscriminatorLoss(nn.Module):
    r"""Base class for all discriminator losses.

    .. note:: All Losses meant to be minimized for optimizing the Discriminator must subclass this.

    Args:
        reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the outputs are averaged over batch size.
            If ``sum`` the elements of the output are summed.
        override_train_ops (function, optional): Function to be used in place of the default ``train_ops``
    """

    def __init__(self, reduction="mean", override_train_ops=None):
        super(DiscriminatorLoss, self).__init__()
        self.reduction = reduction
        self.override_train_ops = override_train_ops
        self.arg_map = {}

    def set_arg_map(self, value):
        r"""Updates the ``arg_map`` for passing a different value to the ``train_ops``.

        Args:
            value (dict): A mapping of the ``argument name`` in the method signature and the
                variable name in the ``Trainer`` it corresponds to.

        .. note::
            If the ``train_ops`` signature is
            ``train_ops(self, gen, disc, optimizer_discriminator, device, batch_size, labels=None)``
            then we need to map ``gen`` to ``generator`` and ``disc`` to ``discriminator``.
            In this case we make the following function call
            ``loss.set_arg_map({"gen": "generator", "disc": "discriminator"})``.
        """
        self.arg_map.update(value)

    def train_ops(
        self,
        generator,
        discriminator,
        optimizer_discriminator,
        real_inputs,
        device,
        labels=None,
    ):
        r"""Defines the standard ``train_ops`` used by most losses. Losses which have a different
        training procedure can either ``subclass`` it **(recommended approach)** or make use of
        ``override_train_ops`` argument.

        The ``standard optimization algorithm`` for the ``discriminator`` defined in this train_ops
        is as follows:

        1. :math:`fake = generator(noise)`
        2. :math:`value_1 = discriminator(fake)`
        3. :math:`value_2 = discriminator(real)`
        4. :math:`loss = loss\_function(value_1, value_2)`
        5. Backpropagate by computing :math:`\nabla loss`
        6. Run a step of the optimizer for discriminator

        Args:
            generator (torchgan.models.Generator): The model to be optimized.
            discriminator (torchgan.models.Discriminator): The discriminator which judges the
                performance of the generator.
            optimizer_discriminator (torch.optim.Optimizer): Optimizer which updates the ``parameters``
                of the ``discriminator``.
            real_inputs (torch.Tensor): The real data to be fed to the ``discriminator``.
            device (torch.device): Device on which the ``generator`` and ``discriminator`` is present.
            batch_size (int): Batch Size of the data infered from the ``DataLoader`` by the ``Trainer``.
            labels (torch.Tensor, optional): Labels for the data.

        Returns:
            Scalar value of the loss.
        """
        if self.override_train_ops is not None:
            return self.override_train_ops(
                self,
                generator,
                discriminator,
                optimizer_discriminator,
                real_inputs,
                device,
                labels,
            )
        else:
            if labels is None and (
                generator.label_type == "required"
                or discriminator.label_type == "required"
            ):
                raise Exception("GAN model requires labels for training")
            batch_size = real_inputs.size(0)
            noise = torch.randn(batch_size, generator.encoding_dims, device=device)
            if generator.label_type == "generated":
                label_gen = torch.randint(
                    0, generator.num_classes, (batch_size,), device=device
                )
            optimizer_discriminator.zero_grad()
            if discriminator.label_type == "none":
                dx = discriminator(real_inputs)
            elif discriminator.label_type == "required":
                dx = discriminator(real_inputs, labels)
            else:
                dx = discriminator(real_inputs, label_gen)
            if generator.label_type == "none":
                fake = generator(noise)
            elif generator.label_type == "required":
                fake = generator(noise, labels)
            else:
                fake = generator(noise, label_gen)
            if discriminator.label_type == "none":
                dgz = discriminator(fake.detach())
            else:
                if generator.label_type == "generated":
                    dgz = discriminator(fake.detach(), label_gen)
                else:
                    dgz = discriminator(fake.detach(), labels)
            loss = self.forward(dx, dgz)
            loss.backward()
            optimizer_discriminator.step()
            # NOTE(avik-pal): This will error if reduction is is 'none'
            return loss.item()



from .auxclassifier import *
from .boundaryequilibrium import *
from .draganpenalty import *
from .energybased import *
from .featurematching import *
from .functional import *
from .historical import *
from .leastsquares import *
from .loss import *
from .minimax import *
from .mutualinfo import *
from .wasserstein import *



import torch
import torch.nn.functional as F
import torchvision

from ..utils import reduce
from .metric import EvaluationMetric

__all__ = ["ClassifierScore"]


class ClassifierScore(EvaluationMetric):
    r"""
    Computes the Classifier Score of a Model. Also popularly known as the Inception Score.
    The ``classifier`` can be any model. It also supports models outside of torchvision models.
    For more details on how to use custom trained models look up the tutorials.

    Args:
        classifier (torch.nn.Module, optional) : The model to be used as a base to compute the classifier
            score. If ``None`` is passed the pretrained ``torchvision.models.inception_v3`` is used.

            .. note ::
                Ensure that the classifier is on the same ``device`` as the Trainer to avoid sudden
                crash.
        transform (torchvision.transforms, optional) : Transformations applied to the image before feeding
            it to the classifier. Look up the documentation of the torchvision models for this transforms.
        sample_size (int): Batch Size for calculation of Classifier Score.
    """

    def __init__(self, classifier=None, transform=None, sample_size=1):
        super(ClassifierScore, self).__init__()
        self.classifier = (
            torchvision.models.inception_v3(True) if classifier is None else classifier
        )
        self.classifier.eval()
        self.transform = transform
        self.sample_size = sample_size

    def preprocess(self, x):
        r"""
        Preprocessor for the Classifier Score. It transforms the image as per the transform requirements
        and feeds it to the classifier.

        Args:
            x (torch.Tensor) : Image in tensor format

        Returns:
            The output from the classifier.
        """
        x = x if self.transform is None else self.transform(x)
        return self.classifier(x)

    def calculate_score(self, x):
        r"""
        Computes the Inception Score for the Input.

        Args:
            x (torch.Tensor) : Image in tensor format

        Returns:
            The Inception Score.
        """
        p = F.softmax(x, dim=1)
        q = torch.mean(p, dim=0)
        kl = torch.sum(p * (F.log_softmax(x, dim=1) - torch.log(q)), dim=1)
        return torch.exp(reduce(kl, "mean")).data

    def metric_ops(self, generator, device):
        r"""Defines the set of operations necessary to compute the ClassifierScore.

        Args:
            generator (torchgan.models.Generator): The generator which needs to be evaluated.
            device (torch.device): Device on which the generator is present.

        Returns:
            The Classifier Score (scalar quantity)
        """
        noise = torch.randn(self.sample_size, generator.encoding_dims, device=device)
        img = generator(noise).detach()
        score = self.__call__(img)
        return score



import torch
import torch.nn.functional as F

from ..utils import reduce
from .loss import DiscriminatorLoss, GeneratorLoss

__all__ = ["FeatureMatchingGeneratorLoss"]


class FeatureMatchingGeneratorLoss(GeneratorLoss):
    r"""Feature Matching Generator loss from
    `"Improved Training of GANs by Salimans et. al." <https://arxiv.org/abs/1606.03498>`_ paper

    The loss can be described as:

    .. math:: L(G) = ||f(x)-f(G(z))||_2

    where

    - :math:`G` : Generator
    - :math:`f` : An intermediate activation from the discriminator
    - :math:`z` : A sample from the noise prior

    Args:
        reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the outputs are averaged over batch size.
            If ``sum`` the elements of the output are summed.
        override_train_ops (function, optional): Function to be used in place of the default ``train_ops``
    """

    def forward(self, fx, fgz):
        r"""Computes the loss for the given input.

        Args:
            dx (torch.Tensor) : Output of the Discriminator with real data. It must have the
                                dimensions (N, \*) where \* means any number of additional
                                dimensions.
            dgz (torch.Tensor) : Output of the Discriminator with generated data. It must have the
                                 dimensions (N, \*) where \* means any number of additional
                                 dimensions.

        Returns:
            scalar if reduction is applied else Tensor with dimensions (N, \*).
        """
        return F.mse_loss(fgz, fx, reduction=self.reduction)

    def train_ops(
        self,
        generator,
        discriminator,
        optimizer_generator,
        real_inputs,
        device,
        labels=None,
    ):
        r"""Defines the standard ``train_ops`` used for feature matching.

        The ``standard optimization algorithm`` for the ``generator`` defined in this train_ops
        is as follows:

        1. :math:`fake = generator(noise)`
        2. :math:`value_1 = discriminator(fake)` where :math:`value_1` is an activation of an intermediate
                                                 discriminator layer
        3. :math:`value_2 = discriminator(real)` where :math:`value_2` is an activation of the same intermediate
                                                 discriminator layer
        4. :math:`loss = loss\_function(value_1, value_2)`
        5. Backpropagate by computing :math:`\nabla loss`
        6. Run a step of the optimizer for generator

        Args:
            generator (torchgan.models.Generator): The model to be optimized.
            discriminator (torchgan.models.Discriminator): The discriminator which judges the
                performance of the generator.
            optimizer_generator (torch.optim.Optimizer): Optimizer which updates the ``parameters``
                of the ``generator``.
            real_inputs (torch.Tensor): The real data to be fed to the ``discriminator``.
            device (torch.device): Device on which the ``generator`` and ``discriminator`` is present.
            labels (torch.Tensor, optional): Labels for the data.

        Returns:
            Scalar value of the loss.
        """
        if self.override_train_ops is not None:
            return self.override_train_ops(
                generator, discriminator, optimizer_generator, device, labels
            )
        else:
            if labels is None and generator.label_type == "required":
                raise Exception("GAN model requires labels for training")
            batch_size = real_inputs.size(0)
            noise = torch.randn(batch_size, generator.encoding_dims, device=device)
            optimizer_generator.zero_grad()
            if generator.label_type == "generated":
                label_gen = torch.randint(
                    0, generator.num_classes, (batch_size,), device=device
                )
            if generator.label_type == "none":
                fake = generator(noise)
            elif generator.label_type == "required":
                fake = generator(noise, labels)
            elif generator.label_type == "generated":
                fake = generator(noise, label_gen)

            if discriminator.label_type == "none":
                fx = discriminator(real_inputs, feature_matching=True)
                fgz = discriminator(fake, feature_matching=True)
            else:
                if discriminator.label_type == "generated":
                    fx = discriminator(real_inputs, label_gen, feature_matching=True)
                else:
                    fx = discriminator(real_inputs, labels, feature_matching=True)
                if generator.label_type == "generated":
                    fgz = discriminator(fake, label_gen, feature_matching=True)
                else:
                    fgz = discriminator(fake, labels, feature_matching=True)
            loss = self.forward(fx, fgz)
            loss.backward()
            optimizer_generator.step()
            return loss.item()



import torch
import torch.nn as nn

__all__ = ["MinibatchDiscrimination1d"]

# The original paper by Salimans et. al. discusses only 1D minibatch discrimination
class MinibatchDiscrimination1d(nn.Module):
    r"""1D Minibatch Discrimination Module as proposed in the paper `"Improved Techniques for
    Training GANs by Salimans et. al." <https://arxiv.org/abs/1805.08318>`_

    Allows the Discriminator to easily detect mode collapse by augmenting the activations to the succeeding
    layer with side information that allows it to determine the 'closeness' of the minibatch examples
    with each other

    .. math :: M_i = T * f(x_{i})
    .. math :: c_b(x_{i}, x_{j}) = \exp(-||M_{i, b} - M_{j, b}||_1) \in \mathbb{R}.
    .. math :: o(x_{i})_b &= \sum_{j=1}^{n} c_b(x_{i},x_{j}) \in \mathbb{R} \\
    .. math :: o(x_{i}) &= \Big[ o(x_{i})_1, o(x_{i})_2, \dots, o(x_{i})_B \Big] \in \mathbb{R}^B \\
    .. math :: o(X) \in \mathbb{R}^{n \times B}

    This is followed by concatenating :math:`o(x_{i})` and :math:`f(x_{i})`

    where

    - :math:`f(x_{i}) \in \mathbb{R}^A` : Activations from an intermediate layer
    - :math:`f(x_{i}) \in \mathbb{R}^A` : Parameter Tensor for generating minibatch discrimination matrix


    Args:
        in_features (int): Features input corresponding to dimension :math:`A`
        out_features (int): Number of output features that are to be concatenated corresponding to dimension :math:`B`
        intermediate_features (int): Intermediate number of features corresponding to dimension :math:`C`

    Returns:
        A Tensor of size :math:`(N, in_features + out_features)` where :math:`N` is the batch size
    """

    def __init__(self, in_features, out_features, intermediate_features=16):
        super(MinibatchDiscrimination1d, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.intermediate_features = intermediate_features

        self.T = nn.Parameter(
            torch.Tensor(in_features, out_features, intermediate_features)
        )
        nn.init.normal_(self.T)

    def forward(self, x):
        r"""Computes the output of the Minibatch Discrimination Layer

        Args:
            x (torch.Tensor): A Torch Tensor of dimensions :math: `(N, infeatures)`

        Returns:
            3D Torch Tensor of size :math: `(N,infeatures + outfeatures)` after applying Minibatch Discrimination
        """
        M = torch.mm(x, self.T.view(self.in_features, -1))
        M = M.view(-1, self.out_features, self.intermediate_features).unsqueeze(0)
        M_t = M.permute(1, 0, 2, 3)
        # Broadcasting reduces the matrix subtraction to the form desired in the paper
        out = torch.sum(torch.exp(-(torch.abs(M - M_t).sum(3))), dim=0) - 1
        return torch.cat([x, out], 1)



import torch
import torch.distributions as distributions
import torch.nn as nn
import torch.nn.functional as F

from .dcgan import DCGANDiscriminator, DCGANGenerator

__all__ = ["InfoGANGenerator", "InfoGANDiscriminator"]


class InfoGANGenerator(DCGANGenerator):
    r"""Generator for InfoGAN based on the Deep Convolutional GAN (DCGAN) architecture, from
    `"InfoGAN : Interpretable Representation Learning With Information Maximizing Generative Aversarial Nets
    by Chen et. al. " <https://arxiv.org/abs/1606.03657>`_ paper

    Args:
        dim_dis (int): Dimension of the discrete latent code sampled from the prior.
        dim_cont (int): Dimension of the continuous latent code sampled from the prior.
        encoding_dims (int, optional): Dimension of the encoding vector sampled from the noise prior.
        out_size (int, optional): Height and width of the input image to be generated. Must be at
            least 16 and should be an exact power of 2.
        out_channels (int, optional): Number of channels in the output Tensor.
        step_channels (int, optional): Number of channels in multiples of which the DCGAN steps up
            the convolutional features. The step up is done as dim :math:`z \rightarrow d \rightarrow
            2 \times d \rightarrow 4 \times d \rightarrow 8 \times d` where :math:`d` = step_channels.
        batchnorm (bool, optional): If True, use batch normalization in the convolutional layers of
            the generator.
        nonlinearity (torch.nn.Module, optional): Nonlinearity to be used in the intermediate
            convolutional layers. Defaults to ``LeakyReLU(0.2)`` when None is passed.
        last_nonlinearity (torch.nn.Module, optional): Nonlinearity to be used in the final
            convolutional layer. Defaults to ``Tanh()`` when None is passed.

    Example:
        >>> import torchgan.models as models
        >>> G = models.InfoGANGenerator(10, 30)
        >>> z = torch.randn(10, 100)
        >>> c_cont = torch.randn(10, 10)
        >>> c_dis = torch.randn(10, 30)
        >>> x = G(z, c_cont, c_dis)
    """

    def __init__(
        self,
        dim_dis,
        dim_cont,
        encoding_dims=100,
        out_size=32,
        out_channels=3,
        step_channels=64,
        batchnorm=True,
        nonlinearity=None,
        last_nonlinearity=None,
    ):
        super(InfoGANGenerator, self).__init__(
            encoding_dims + dim_dis + dim_cont,
            out_size,
            out_channels,
            step_channels,
            batchnorm,
            nonlinearity,
            last_nonlinearity,
        )
        self.encoding_dims = encoding_dims
        self.dim_cont = dim_cont
        self.dim_dis = dim_dis

    def forward(self, z, c_dis=None, c_cont=None):
        z_cat = (
            torch.cat([z, c_dis, c_cont], dim=1)
            if c_dis is not None and c_cont is not None
            else z
        )
        return super(InfoGANGenerator, self).forward(z_cat)


class InfoGANDiscriminator(DCGANDiscriminator):
    r"""Discriminator for InfoGAN based on the Deep Convolutional GAN (DCGAN) architecture, from
    `"InfoGAN : Interpretable Representation Learning With Information Maximizing Generative Aversarial Nets
    by Chen et. al. " <https://arxiv.org/abs/1606.03657>`_ paper

    The approximate conditional probability distribution over the latent code Q(c|x) is chosen to be a factored
    Gaussian for the continuous latent code and a Categorical distribution for the discrete latent code

    Args:
        dim_dis (int): Dimension of the discrete latent code sampled from the prior.
        dim_cont (int): Dimension of the continuous latent code sampled from the prior.
        encoding_dims (int, optional): Dimension of the encoding vector sampled from the noise prior.
        in_size (int, optional): Height and width of the input image to be evaluated. Must be at
            least 16 and should be an exact power of 2.
        in_channels (int, optional): Number of channels in the input Tensor.
        step_channels (int, optional): Number of channels in multiples of which the DCGAN steps up
            the convolutional features. The step up is done as dim :math:`z \rightarrow d \rightarrow
            2 \times d \rightarrow 4 \times d \rightarrow 8 \times d` where :math:`d` = step_channels.
        batchnorm (bool, optional): If True, use batch normalization in the convolutional layers of
            the generator.
        nonlinearity (torch.nn.Module, optional): Nonlinearity to be used in the intermediate
            convolutional layers. Defaults to ``LeakyReLU(0.2)`` when None is passed.
        last_nonlinearity (torch.nn.Module, optional): Nonlinearity to be used in the final
            convolutional layer. Defaults to ``Tanh()`` when None is passed.
        latent_nonlinearity (torch.nn.Module, optional): Nonlinearity to be used in the ``dist_conv``.
            Defaults to ``LeakyReLU(0.2)`` when None is passed.
    Example:
        >>> import torchgan.models as models
        >>> D = models.InfoGANDiscriminator(10, 30)
        >>> x = torch.randn(10, 3, 32, 32)
        >>> score, q_categorical, q_gaussian = D(x, return_latents=True)
    """

    def __init__(
        self,
        dim_dis,
        dim_cont,
        in_size=32,
        in_channels=3,
        step_channels=64,
        batchnorm=True,
        nonlinearity=None,
        last_nonlinearity=None,
        latent_nonlinearity=None,
    ):
        self.dim_cont = dim_cont
        self.dim_dis = dim_dis
        super(InfoGANDiscriminator, self).__init__(
            in_size,
            in_channels,
            step_channels,
            batchnorm,
            nonlinearity,
            last_nonlinearity,
        )

        self.latent_nl = (
            nn.LeakyReLU(0.2) if latent_nonlinearity is None else latent_nonlinearity
        )
        d = self.n * 2 ** (in_size.bit_length() - 4)
        if batchnorm is True:
            self.dist_conv = nn.Sequential(
                nn.Conv2d(d, d, 4, 1, 0, bias=not batchnorm),
                nn.BatchNorm2d(d),
                self.latent_nl,
            )
        else:
            self.dist_conv = nn.Sequential(
                nn.Conv2d(d, d, 4, 1, 0, bias=not batchnorm), self.latent_nl
            )

        self.dis_categorical = nn.Linear(d, self.dim_dis)

        self.cont_mean = nn.Linear(d, self.dim_cont)
        self.cont_logvar = nn.Linear(d, self.dim_cont)

    def forward(self, x, return_latents=False, feature_matching=False):
        x = self.model(x)
        if feature_matching is True:
            return x
        critic_score = self.disc(x)
        x = self.dist_conv(x).view(-1, x.size(1))
        dist_dis = distributions.OneHotCategorical(logits=self.dis_categorical(x))
        dist_cont = distributions.Normal(
            loc=self.cont_mean(x), scale=torch.exp(0.5 * self.cont_logvar(x))
        )
        return (
            critic_score,
            dist_dis,
            dist_cont if return_latents is True else critic_score,
        )



import torch

from .functional import (
    boundary_equilibrium_discriminator_loss,
    boundary_equilibrium_generator_loss,
)
from .loss import DiscriminatorLoss, GeneratorLoss

__all__ = ["BoundaryEquilibriumGeneratorLoss", "BoundaryEquilibriumDiscriminatorLoss"]


class BoundaryEquilibriumGeneratorLoss(GeneratorLoss):
    r"""Boundary Equilibrium GAN generator loss from
    `"BEGAN : Boundary Equilibrium Generative Adversarial Networks
    by Berthelot et. al." <https://arxiv.org/abs/1703.10717>`_ paper

    The loss can be described as

    .. math:: L(G) = D(G(z))

    where

    - :math:`G` : Generator
    - :math:`D` : Discriminator

    Args:
        reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the outputs are averaged over batch size.
            If ``sum`` the elements of the output are summed.
        override_train_ops (function, optional): Function to be used in place of the default ``train_ops``
    """

    def forward(self, dgz):
        r"""Computes the loss for the given input.

        Args:
            dgz (torch.Tensor) : Output of the Discriminator with generated data. It must have the
                                 dimensions (N, \*) where \* means any number of additional
                                 dimensions.

        Returns:
            scalar if reduction is applied else Tensor with dimensions (N, \*).
        """
        return boundary_equilibrium_generator_loss(dgz, self.reduction)


class BoundaryEquilibriumDiscriminatorLoss(DiscriminatorLoss):
    r"""Boundary Equilibrium GAN discriminator loss from
    `"BEGAN : Boundary Equilibrium Generative Adversarial Networks
    by Berthelot et. al." <https://arxiv.org/abs/1703.10717>`_ paper

    The loss can be described as

    .. math:: L(D) = D(x) - k_t \times D(G(z))

    .. math:: k_{t+1} = k_t + \lambda \times (\gamma \times D(x) - D(G(z)))

    where

    - :math:`G` : Generator
    - :math:`D` : Discriminator
    - :math:`k_t` : Running average of the balance point of G and D
    - :math:`\lambda` : Learning rate of the running average
    - :math:`\gamma` : Goal bias hyperparameter

    Args:
        reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the outputs are averaged over batch size.
            If ``sum`` the elements of the output are summed.
        override_train_ops (function, optional): Function to be used in place ofthe default ``train_ops``
        init_k (float, optional): Initial value of the balance point ``k``.
        lambd (float, optional): Learning rate of the running average.
        gamma (float, optional): Goal bias hyperparameter.
    """

    def __init__(
        self,
        reduction="mean",
        override_train_ops=None,
        init_k=0.0,
        lambd=0.001,
        gamma=0.75,
    ):
        super(BoundaryEquilibriumDiscriminatorLoss, self).__init__(
            reduction, override_train_ops
        )
        self.reduction = reduction
        self.override_train_ops = override_train_ops
        self.k = init_k
        self.lambd = lambd
        self.gamma = gamma
        # TODO(Aniket1998): Integrate this with the metrics API in a later release
        self.convergence_metric = None

    def forward(self, dx, dgz):
        r"""Computes the loss for the given input.

        Args:
            dx (torch.Tensor) : Output of the Discriminator with real data. It must have the
                                dimensions (N, \*) where \* means any number of additional
                                dimensions.
            dgz (torch.Tensor) : Output of the Discriminator with generated data. It must have the
                                 dimensions (N, \*) where \* means any number of additional
                                 dimensions.

        Returns:
            A tuple of 3 loss values, namely the ``total loss``, ``loss due to real data`` and ``loss
            due to fake data``.
        """
        return boundary_equilibrium_discriminator_loss(dx, dgz, self.k, self.reduction)

    def set_k(self, k=0.0):
        r"""Change the default value of k

        Args:
            k (float, optional) : New value to be set.
        """
        self.k = k

    def update_k(self, loss_real, loss_fake):
        r"""Update the running mean of k for each forward pass.

        The update takes place as

        .. math:: k_{t+1} = k_t + \lambda \times (\gamma \times D(x) - D(G(z)))

        Args:
            loss_real (float): :math:`D(x)`
            loss_fake (float): :math:`D(G(z))`
        """
        diff = self.gamma * loss_real - loss_fake
        self.k += self.lambd * diff
        # TODO(Aniket1998): Develop this into a proper TorchGAN convergence metric
        self.convergence_metric = loss_real + abs(diff)
        self.k = max(min(self.k, 1.0), 0.0)

    def train_ops(
        self,
        generator,
        discriminator,
        optimizer_discriminator,
        real_inputs,
        device,
        labels=None,
    ):
        r"""Defines the standard ``train_ops`` used by boundary equilibrium loss.

        The ``standard optimization algorithm`` for the ``discriminator`` defined in this train_ops
        is as follows:

        1. :math:`fake = generator(noise)`
        2. :math:`value_1 = discriminator(fake)`
        3. :math:`value_2 = discriminator(real)`
        4. :math:`loss = loss\_function(value_1, value_2)`
        5. Backpropagate by computing :math:`\nabla loss`
        6. Run a step of the optimizer for discriminator
        7. Update the value of :math: `k`.

        Args:
            generator (torchgan.models.Generator): The model to be optimized.
            discriminator (torchgan.models.Discriminator): The discriminator which judges the
                performance of the generator.
            optimizer_discriminator (torch.optim.Optimizer): Optimizer which updates the ``parameters``
                of the ``discriminator``.
            real_inputs (torch.Tensor): The real data to be fed to the ``discriminator``.
            device (torch.device): Device on which the ``generator`` and ``discriminator`` is present.
            labels (torch.Tensor, optional): Labels for the data.

        Returns:
            Scalar value of the loss.
        """
        if self.override_train_ops is not None:
            return self.override_train_ops(
                self,
                generator,
                discriminator,
                optimizer_discriminator,
                real_inputs,
                device,
                labels,
            )
        else:
            if labels is None and (
                generator.label_type == "required"
                or discriminator.label_type == "required"
            ):
                raise Exception("GAN model requires labels for training")
            batch_size = real_inputs.size(0)
            noise = torch.randn(batch_size, generator.encoding_dims, device=device)
            if generator.label_type == "generated":
                label_gen = torch.randint(
                    0, generator.num_classes, (batch_size,), device=device
                )
            optimizer_discriminator.zero_grad()
            if discriminator.label_type == "none":
                dx = discriminator(real_inputs)
            elif discriminator.label_type == "required":
                dx = discriminator(real_inputs, labels)
            else:
                dx = discriminator(real_inputs, label_gen)
            if generator.label_type == "none":
                fake = generator(noise)
            elif generator.label_type == "required":
                fake = generator(noise, labels)
            else:
                fake = generator(noise, label_gen)
            if discriminator.label_type == "none":
                dgz = discriminator(fake.detach())
            else:
                if generator.label_type == "generated":
                    dgz = discriminator(fake.detach(), label_gen)
                else:
                    dgz = discriminator(fake.detach(), labels)
            loss_total, loss_real, loss_fake = self.forward(dx, dgz)
            loss_total.backward()
            optimizer_discriminator.step()
            self.update_k(loss_real.item(), loss_fake.item())
            return loss_total.item()



from .backends import *
from .visualize import *

if TENSORBOARD_LOGGING == 1:
    from tensorboardX import SummaryWriter
if VISDOM_LOGGING == 1:
    import visdom

__all__ = ["Logger"]


class Logger(object):
    r"""Base Logger class. It controls the executions of all the Visualizers and is deeply
    integrated with the functioning of the Trainer.

    .. note::
        The ``Logger`` has been designed to be controlled internally by the ``Trainer``. It is
        recommended that the user does not attempt to use it externally in any form.

    .. warning::
        This ``Logger`` is meant to work on the standard Visualizers available. Work is being
        done to support custom Visualizers in a clean way. But currently it is not possible to
        do so.

    Args:
        trainer (torchgan.trainer.Trainer): The base trainer used for training.
        losses_list (list): A list of the Loss Functions that need to be minimized. For a list of
            pre-defined losses look at :mod:`torchgan.losses`. All losses in the list must be a
            subclass of atleast ``GeneratorLoss`` or ``DiscriminatorLoss``.
        metrics_list (list, optional): List of Metric Functions that need to be logged. For a list of
            pre-defined metrics look at :mod:`torchgan.metrics`. All losses in the list must be a
            subclass of ``EvaluationMetric``.
        visdom_port (int, optional): Port to log using ``visdom``. A deafult server is started
            at port ``8097``. So manually a new server has to be started if the post is changed.
            This is ignored if ``VISDOM_LOGGING`` is ``0``.
        log_dir (str, optional): Directory where TensorboardX should store the logs. This is
            ignored if ``TENSORBOARD_LOGGING`` is ``0``.
        writer (tensorboardX.SummaryWriter, optonal): Send a `SummaryWriter` if you
            don't want to start a new SummaryWriter.
        test_noise (torch.Tensor, optional): If provided then it will be used as the noise for image
            sampling.
        nrow (int, optional): Number of rows in which the image is to be stored.
    """

    def __init__(
        self,
        trainer,
        losses_list,
        metrics_list=None,
        visdom_port=8097,
        log_dir=None,
        writer=None,
        nrow=8,
        test_noise=None,
    ):
        if TENSORBOARD_LOGGING == 1:
            self.writer = SummaryWriter(log_dir) if writer is None else writer
        else:
            self.writer = None
        self.logger_end_epoch = []
        self.logger_mid_epoch = []
        self.logger_end_epoch.append(
            ImageVisualize(
                trainer, writer=self.writer, test_noise=test_noise, nrow=nrow
            )
        )
        self.logger_mid_epoch.append(
            GradientVisualize(trainer.model_names, writer=self.writer)
        )
        if metrics_list is not None:
            self.logger_end_epoch.append(
                MetricVisualize(metrics_list, writer=self.writer)
            )
        self.logger_mid_epoch.append(LossVisualize(losses_list, writer=self.writer))

    def get_loss_viz(self):
        r"""Get the LossVisualize object.
        """
        return self.logger_mid_epoch[1]

    def get_metric_viz(self):
        r"""Get the MetricVisualize object.
        """
        return self.logger_end_epoch[1]

    def get_grad_viz(self):
        r"""Get the GradientVisualize object.
        """
        return self.logger_mid_epoch[0]

    def register(self, visualize, *args, mid_epoch=True, **kwargs):
        r"""Register a new ``Visualize`` object with the Logger.

        Args:
            visualize (torchgan.logging.Visualize): Class name of the visualizer.
            mid_epoch (bool, optional): Set it to ``False`` if it is to be executed once the epoch is
                over. Otherwise it is executed after every call to the ``train_iter``.
        """
        if mid_epoch:
            self.logger_mid_epoch.append(visualize(*args, writer=self.writer, **kwargs))
        else:
            self.logger_end_epoch.append(visualize(*args, writer=self.writer, **kwargs))

    def close(self):
        r"""Turns off the tensorboard ``SummaryWriter`` if it were created.
        """
        if self.writer is not None:
            self.writer.close()

    def run_mid_epoch(self, trainer, *args):
        r"""Runs the Visualizers after every call to the ``train_iter``.

        Args:
            trainer (torchgan.trainer.Trainer): The base trainer used for training.
        """
        for logger in self.logger_mid_epoch:
            if (
                type(logger).__name__ == "LossVisualize"
                or type(logger).__name__ == "GradientVisualize"
            ):
                logger(trainer, lock_console=True)
            else:
                logger(*args, lock_console=True)

    def run_end_epoch(self, trainer, epoch, time_duration, *args):
        r"""Runs the Visualizers at the end of one epoch.

        Args:
            trainer (torchgan.trainer.Trainer): The base trainer used for training.
            epoch (int): The epoch number which was completed.
        """
        print("Epoch {} Summary".format(epoch + 1))
        print("Epoch time duration : {}".format(time_duration))
        for logger in self.logger_mid_epoch:
            if type(logger).__name__ == "LossVisualize":
                logger(trainer)
            elif type(logger).__name__ == "GradientVisualize":
                logger.report_end_epoch()
            else:
                logger(*args)
        for logger in self.logger_end_epoch:
            if type(logger).__name__ == "ImageVisualize":
                logger(trainer)
            elif type(logger).__name__ == "MetricVisualize":
                logger()
            else:
                logger(*args)
        print()



#!/usr/bin/env python
import io
import os
import re
import shutil
import sys

from pkg_resources import DistributionNotFound, get_distribution
from setuptools import find_packages, setup


def read(*names, **kwargs):
    with io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8"),
    ) as fp:
        return fp.read()


def get_dist(pkgname):
    try:
        return get_distribution(pkgname)
    except DistributionNotFound:
        return None


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


PATH_ROOT = os.path.dirname(__file__)
VERSION = find_version("torchgan", "__init__.py")


def load_requirements(path_dir=PATH_ROOT, comment_char='#'):
    with open(os.path.join(path_dir, 'requirements.txt'), 'r') as file:
        lines = [ln.strip() for ln in file.readlines()]
    reqs = []
    for ln in lines:
        # filer all comments
        if comment_char in ln:
            ln = ln[:ln.index(comment_char)]
        if ln:  # if requirement is not empty
            reqs.append(ln)
    return reqs

setup(
    # Metadata
    name="torchgan",
    version=VERSION,
    author="Avik Pal & Aniket Das",
    author_email="avikpal@cse.iitk.ac.in",
    url="https://github.com/torchgan/torchgan",
    description="Research Framework for easy and efficient training of GANs based on Pytorch",
    license="MIT",
    # Package info
    packages=find_packages(exclude=("test",)),
    zip_safe=True,
    install_requires=load_requirements(PATH_ROOT),
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords=['deep learning', 'pytorch', 'GAN', 'AI'],
    python_requires='>=3.6',
    classifiers=[
        'Environment :: Console',
        'Natural Language :: English',
        # How mature is this project? Common values are
        #   3 - Alpha, 4 - Beta, 5 - Production/Stable
        'Development Status :: 4 - Beta',
        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Image Recognition',
        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



import torch
import torch.nn as nn
from torch.nn import Parameter

__all__ = ["SpectralNorm2d"]

# NOTE(avik-pal): This code has been adapted from
#                 https://github.com/heykeetae/Self-Attention-GAN/blob/master/spectral.py
class SpectralNorm2d(nn.Module):
    r"""2D Spectral Norm Module as described in `"Spectral Normalization
    for Generative Adversarial Networks by Miyato et. al." <https://arxiv.org/abs/1802.05957>`_
    The spectral norm is computed using ``power iterations``.

    Computation Steps:

    .. math:: v_{t + 1} = \frac{W^T W v_t}{||W^T W v_t||} = \frac{(W^T W)^t v}{||(W^T W)^t v||}
    .. math:: u_{t + 1} = W v_t
    .. math:: v_{t + 1} = W^T u_{t + 1}
    .. math:: Norm(W) = ||W v|| = u^T W v
    .. math:: Output = \frac{W}{Norm(W)} = \frac{W}{u^T W v}

    Args:
        module (torch.nn.Module): The Module on which the Spectral Normalization needs to be
            applied.
        name (str, optional): The attribute of the ``module`` on which normalization needs to
            be performed.
        power_iterations (int, optional): Total number of iterations for the norm to converge.
            ``1`` is usually enough given the weights vary quite gradually.

    Example:
        .. code:: python

            >>> layer = SpectralNorm2d(Conv2d(3, 16, 1))
            >>> x = torch.rand(1, 3, 10, 10)
            >>> layer(x)
    """

    def __init__(self, module, name="weight", power_iterations=1):
        super(SpectralNorm2d, self).__init__()
        self.module = module
        self.name = name
        self.power_iterations = power_iterations
        w = getattr(self.module, self.name)
        height = w.data.shape[0]
        width = w.view(height, -1).data.shape[1]
        self.u = Parameter(w.data.new(height).normal_(0, 1), requires_grad=False)
        self.v = Parameter(w.data.new(width).normal_(0, 1), requires_grad=False)
        self.u.data = self._l2normalize(self.u.data)
        self.v.data = self._l2normalize(self.v.data)
        self.w_bar = Parameter(w.data)
        del self.module._parameters[self.name]

    def _l2normalize(self, x, eps=1e-12):
        r"""Function to calculate the ``L2 Normalized`` form of a Tensor

        Args:
            x (torch.Tensor): Tensor which needs to be normalized.
            eps (float, optional): A small value needed to avoid infinite values.

        Returns:
            Normalized form of the tensor ``x``.
        """
        return x / (torch.norm(x) + eps)

    def forward(self, *args):
        r"""Computes the output of the ``module`` and appies spectral normalization to the
        ``name`` attribute of the ``module``.

        Returns:
            The output of the ``module``.
        """
        height = self.w_bar.data.shape[0]
        for _ in range(self.power_iterations):
            self.v.data = self._l2normalize(
                torch.mv(torch.t(self.w_bar.view(height, -1)), self.u)
            )
            self.u.data = self._l2normalize(
                torch.mv(self.w_bar.view(height, -1), self.v)
            )
        sigma = self.u.dot(self.w_bar.view(height, -1).mv(self.v))
        setattr(self.module, self.name, self.w_bar / sigma.expand_as(self.w_bar))
        return self.module.forward(*args)



import os
import sys
import unittest

import torch
import torch.distributions as ds
from torchgan.losses import *

sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))


class TestLosses(unittest.TestCase):
    def match_losses(
        self,
        l_g,
        l_d,
        dx,
        dgz,
        gen_loss_mean,
        gen_loss_sum,
        gen_loss_none,
        d_loss_mean,
        d_loss_sum,
        d_loss_none,
    ):
        D_X = torch.Tensor(dx).view(-1, 1)
        D_GZ = torch.Tensor(dgz).view(-1, 1)

        self.assertAlmostEqual(d_loss_mean, l_d(D_X, D_GZ).item(), places=5)
        l_d.reduction = "sum"
        self.assertAlmostEqual(d_loss_sum, l_d(D_X, D_GZ).item(), places=5)
        l_d.reduction = "none"
        loss_none = l_d(D_X, D_GZ).view(-1, 1)
        for i in range(4):
            self.assertAlmostEqual(d_loss_none[i], loss_none[i].item(), places=5)

        self.assertAlmostEqual(gen_loss_mean, l_g(D_GZ).item(), places=5)
        l_g.reduction = "sum"
        self.assertAlmostEqual(gen_loss_sum, l_g(D_GZ).item(), places=5)
        l_g.reduction = "none"
        loss_none = l_g(D_GZ).view(-1, 1)
        for i in range(4):
            self.assertAlmostEqual(gen_loss_none[i], loss_none[i].item(), places=5)

    def test_wasserstein_loss(self):
        dx = [1.3, 2.9, 8.4, 6.3]
        dgz = [4.8, 1.2, -3.5, 5.9]

        gen_loss_mean = -2.1
        gen_loss_sum = -8.4
        gen_loss_none = [-4.8, -1.2, 3.5, -5.9]

        d_loss_mean = -2.625
        d_loss_sum = -10.5
        d_loss_none = [3.5000002, -1.7, -11.9, -0.4000001]

        w_g = WassersteinGeneratorLoss()
        w_d = WassersteinDiscriminatorLoss()
        self.match_losses(
            w_g,
            w_d,
            dx,
            dgz,
            gen_loss_mean,
            gen_loss_sum,
            gen_loss_none,
            d_loss_mean,
            d_loss_sum,
            d_loss_none,
        )

    def test_lsgan_loss(self):
        dx = [1.3, 2.9, 8.4, 6.3]
        dgz = [4.8, 1.2, -3.5, 5.9]

        gen_loss_mean = 7.3425007
        gen_loss_sum = 29.370003
        gen_loss_none = [7.2200007, 0.02000001, 10.125, 12.005]

        d_loss_mean = 19.76125
        d_loss_sum = 79.045
        d_loss_none = [11.565001, 2.525, 33.504997, 31.45]

        l_g = LeastSquaresGeneratorLoss()
        l_d = LeastSquaresDiscriminatorLoss()
        self.match_losses(
            l_g,
            l_d,
            dx,
            dgz,
            gen_loss_mean,
            gen_loss_sum,
            gen_loss_none,
            d_loss_mean,
            d_loss_sum,
            d_loss_none,
        )

    def test_minimax_loss(self):
        dx = [1.3, 2.9, 8.4, 6.3]
        dgz = [4.8, 1.2, -3.5, 5.9]

        factor = -torch.log(torch.sigmoid(torch.ones(1))).item()

        gen_loss_mean = -3.3642528 + factor
        gen_loss_sum = -13.457011 + 4 * factor
        gen_loss_none = [
            -5.1214576 + factor,
            -1.7765441 + factor,
            -0.3430121 + factor,
            -6.215997 + factor,
        ]

        d_loss_mean = 3.1251488
        d_loss_sum = 12.500595
        d_loss_none = [5.0492043, 1.5168452, 0.02997526, 5.90457]

        l_g = MinimaxGeneratorLoss(nonsaturating=False)
        l_d = MinimaxDiscriminatorLoss()
        self.match_losses(
            l_g,
            l_d,
            dx,
            dgz,
            gen_loss_mean,
            gen_loss_sum,
            gen_loss_none,
            d_loss_mean,
            d_loss_sum,
            d_loss_none,
        )

    def test_minimax_nonsaturating_loss(self):
        dx = [1.3, 2.9, 8.4, 6.3]
        dgz = [4.8, 1.2, -3.5, 5.9]

        gen_loss_mean = 0.9509911
        gen_loss_sum = 3.8039644
        gen_loss_none = [8.1960661e-03, 2.6328245e-01, 3.5297503e00, 2.7356991e-03]

        d_loss_mean = 3.1251488
        d_loss_sum = 12.500595
        d_loss_none = [5.0492043, 1.5168452, 0.02997526, 5.90457]

        l_g = MinimaxGeneratorLoss(nonsaturating=True)
        l_d = MinimaxDiscriminatorLoss()
        self.match_losses(
            l_g,
            l_d,
            dx,
            dgz,
            gen_loss_mean,
            gen_loss_sum,
            gen_loss_none,
            d_loss_mean,
            d_loss_sum,
            d_loss_none,
        )

    def test_mutual_info_penalty(self):
        real_loss_mean = 2.600133
        real_loss_sum = 5.200266
        real_losses = [0.7086121, 4.491654]
        mean = torch.Tensor([[1.3, 4.6, 7.1], [0.2, 11.4, 1.0]])
        std = torch.Tensor([[1.0, 0.5, 3.1], [0.2, 3.5, 4.9]])
        logits = torch.Tensor([[0.5, 0.5], [0.75, 0.25]])

        c_dis = torch.Tensor([[0, 1], [1, 0]])
        c_cont = torch.Tensor([[1.4, 4.0, 5.0], [-1.0, 7.0, 2.0]])

        q_cont = ds.Normal(loc=mean, scale=std)
        q_cat = ds.Categorical(logits=logits)

        mutualinfo = MutualInformationPenalty()
        loss_mean = mutualinfo(c_dis, c_cont, q_cat, q_cont)
        self.assertAlmostEqual(loss_mean.item(), real_loss_mean, 5)

        mutualinfo.reduction = "sum"
        loss_sum = mutualinfo(c_dis, c_cont, q_cat, q_cont)
        self.assertAlmostEqual(loss_sum.item(), real_loss_sum, 5)

        mutualinfo.reduction = "none"
        loss = mutualinfo(c_dis, c_cont, q_cat, q_cont)
        for i in range(2):
            self.assertAlmostEqual(loss[i].item(), real_losses[i], 5)



import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = ["SelfAttention2d"]


class SelfAttention2d(nn.Module):
    r"""Self Attention Module as proposed in the paper `"Self-Attention Generative Adversarial
    Networks by Han Zhang et. al." <https://arxiv.org/abs/1805.08318>`_

    .. math:: attention = softmax((query(x))^T * key(x))
    .. math:: output = \gamma * value(x) * attention + x

    where

    - :math:`query` : 2D Convolution Operation
    - :math:`key` : 2D Convolution Operation
    - :math:`value` : 2D Convolution Operation
    - :math:`x` : Input

    Args:
        input_dims (int): The input channel dimension in the input ``x``.
        output_dims (int, optional): The output channel dimension. If ``None`` the output
            channel value is computed as ``input_dims // 8``. So if the ``input_dims`` is **less
            than 8** then the layer will give an error.
        return_attn (bool, optional): Set it to ``True`` if you want the attention values to be
            returned.
    """

    def __init__(self, input_dims, output_dims=None, return_attn=False):
        output_dims = input_dims // 8 if output_dims is None else output_dims
        if output_dims == 0:
            raise Exception(
                "The output dims corresponding to the input dims is 0. Increase the input\
                            dims to 8 or more. Else specify output_dims"
            )
        super(SelfAttention2d, self).__init__()
        self.query = nn.Conv2d(input_dims, output_dims, 1)
        self.key = nn.Conv2d(input_dims, output_dims, 1)
        self.value = nn.Conv2d(input_dims, input_dims, 1)
        self.gamma = nn.Parameter(torch.zeros(1))
        self.return_attn = return_attn

    def forward(self, x):
        r"""Computes the output of the Self Attention Layer

        Args:
            x (torch.Tensor): A 4D Tensor with the channel dimension same as ``input_dims``.

        Returns:
            A tuple of the ``output`` and the ``attention`` if ``return_attn`` is set to ``True``
            else just the ``output`` tensor.
        """
        dims = (x.size(0), -1, x.size(2) * x.size(3))
        out_query = self.query(x).view(dims)
        out_key = self.key(x).view(dims).permute(0, 2, 1)
        attn = F.softmax(torch.bmm(out_key, out_query), dim=-1)
        out_value = self.value(x).view(dims)
        out_value = torch.bmm(out_value, attn).view(x.size())
        out = self.gamma * out_value + x
        if self.return_attn:
            return out, attn
        return out



__all__ = ["EvaluationMetric"]


class EvaluationMetric(object):
    r"""
    Base class for all Evaluation Metrics
    """

    def __init__(self):
        self.arg_map = {}

    def set_arg_map(self, value):
        r"""Updates the ``arg_map`` for passing a different value to the ``metric_ops``.

        Args:
            value (dict): A mapping of the ``argument name`` in the method signature and the
                variable name in the ``Trainer`` it corresponds to.

        .. note::
            If the ``metric_ops`` signature is
            ``metric_ops(self, gen, disc)``
            then we need to map ``gen`` to ``generator`` and ``disc`` to ``discriminator``.
            In this case we make the following function call
            ``metric.set_arg_map({"gen": "generator", "disc": "discriminator"})``.
        """
        self.arg_map.update(value)

    def preprocess(self, x):
        r"""
        Subclasses must override this function and provide their own preprocessing
        pipeline.

        :raises NotImplementedError: If the subclass doesn't override this function.
        """
        raise NotImplementedError

    def calculate_score(self, x):
        r"""
        Subclasses must override this function and provide their own score calculation.

        :raises NotImplementedError: If the subclass doesn't override this function.
        """
        raise NotImplementedError

    def metric_ops(self, generator, discriminator, **kwargs):
        r"""
        Subclasses must override this function and provide their own metric evaluation ops.

        :raises NotImplementedError: If the subclass doesn't override this function.
        """
        raise NotImplementedError

    def __call__(self, x):
        return self.calculate_score(self.preprocess(x))



import torch
import torch.nn as nn

__all__ = ["Generator", "Discriminator"]


class Generator(nn.Module):
    r"""Base class for all Generator models. All Generator models must subclass this.

    Args:
        encoding_dims (int): Dimensions of the sample from the noise prior.
        label_type (str, optional): The type of labels expected by the Generator. The available
            choices are 'none' if no label is needed, 'required' if the original labels are
            needed and 'generated' if labels are to be sampled from a distribution.
    """
    # FIXME(Aniket1998): If a user is overriding the default initializer, he must also
    # override the constructor. Find an efficient workaround by fixing the initialization mechanism
    def __init__(self, encoding_dims, label_type="none"):
        super(Generator, self).__init__()
        self.encoding_dims = encoding_dims
        self.label_type = label_type

    # TODO(Aniket1998): Think of better dictionary lookup based approaches to initialization
    # That allows easy and customizable weight initialization without overriding
    def _weight_initializer(self):
        r"""Default weight initializer for all generator models.
        Models that require custom weight initialization can override this method
        """
        for m in self.modules():
            if isinstance(m, nn.ConvTranspose2d):
                nn.init.kaiming_normal_(m.weight)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1.0)
                nn.init.constant_(m.bias, 0.0)

    def sampler(self, sample_size, device):
        r"""Function to allow sampling data at inference time. Models requiring
        input in any other format must override it in the subclass.

        Args:
            sample_size (int): The number of images to be generated
            device (torch.device): The device on which the data must be generated

        Returns:
            A list of the items required as input
        """
        return [torch.randn(sample_size, self.encoding_dims, device=device)]


class Discriminator(nn.Module):
    r"""Base class for all Discriminator models. All Discriminator models must subclass this.

    Args:
        input_dims (int): Dimensions of the input.
        label_type (str, optional): The type of labels expected by the Discriminator. The available
            choices are 'none' if no label is needed, 'required' if the original labels are
            needed and 'generated' if labels are to be sampled from a distribution.
    """

    def __init__(self, input_dims, label_type="none"):
        super(Discriminator, self).__init__()
        self.input_dims = input_dims
        self.label_type = label_type

    # TODO(Aniket1998): Think of better dictionary lookup based approaches to initialization
    # That allows easy and customizable weight initialization without overriding
    def _weight_initializer(self):
        r"""Default weight initializer for all disciminator models.
        Models that require custom weight initialization can override this method
        """
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1.0)
                nn.init.constant_(m.bias, 0.0)



import os
import sys
import unittest

import torch
from torchgan.metrics import *

sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))


class TestMetrics(unittest.TestCase):
    def test_inception_score(self):
        inception_score = ClassifierScore()
        x = torch.Tensor([[1.0, 2.0, 3.0], [-1.0, 5.0, 3.1]])
        self.assertAlmostEqual(inception_score.calculate_score(x).item(), 1.24357, 4)



import os
import subprocess

from ..utils import getenv_defaults

# Backends available for Visualization

# Tensorboard
TENSORBOARD_LOGGING = int(
    os.getenv("TENSORBOARD_LOGGING", getenv_defaults("tensorboardX"))
)
if TENSORBOARD_LOGGING == 1 and getenv_defaults("tensorboardX") == 0:
    raise Exception(
        "TensorboardX is not installed. Install it or set TENSORBOARD_LOGGING to 0"
    )

# Console
CONSOLE_LOGGING = int(os.getenv("CONSOLE_LOGGING", 1))

# Visdom
VISDOM_LOGGING = int(os.getenv("VISDOM_LOGGING", getenv_defaults("visdom")))
if VISDOM_LOGGING == 1:
    if getenv_defaults("visdom") == 0:
        raise Exception(
            "Visdom is not installed. Install it or set VISDOM_LOGGING to 0"
        )
